import { fetch } from 'whatwg-fetch';
import 'abortcontroller-polyfill/dist/polyfill-patch-fetch';

abstract class Fetch {
  abstract get(url: string, data: any, options: IOptions): Promise<any>;
  abstract post(url: string, data: any, options: IOptions): Promise<any>;
  abstract getData(url: string, data: any, options: IOptions): Promise<any>;
  abstract postData(url: string, data: any, options: IOptions): Promise<any>;
  abstract postFormData(url: string, data: any, options: IOptions): Promise<any>;
  abstract createRequest(url: string, data: any, options: IOptions): Promise<any>;
}

type Tcredentials = 'omit' | 'same-origin' | 'include';
type TMethods =
  | 'GET'
  | 'POST'
  | 'DELETE'
  | 'PUT'
  | 'OPTIONS'
  | 'PATCH'
  | 'TRACE'
  | 'CONNECT'
  | 'HEAD';

type TDataType = 'json' | null;
interface IOptions {
  /**
   * 请求方法
   */
  method?: TMethods;
  /**
   * fetch 的 mode
   */
  mode?: string;
  /**
   * 请求的头信息
   */
  headers?: any;
  /**
   * 请求的 credentials，如 omit、same-origin 或者 include。为了在当前域名内自动发送 cookie
   */
  credentials?: Tcredentials;
  /**
   * 是否进行接口数据缓存
   */
  cache?: boolean;
  /**
   * 是否支持可abort取消请求
   */
  canAbort?: boolean;
  /**
   * AbortContronller的signal
   */
  signal?: any;
  /**
   * 请求的 body 信息
   */
  body?: any;
  /**
   * 请求超时时间ms
   */
  timeout?: number;
  /**
   * 去除headers中的ContentType
   */
  noContentType?: boolean;
  /**
   * 传递的数据格式
   */
  dataType?: TDataType;
}

interface IConstructor {
  defalutParams?: any;
  defaultOptions?: IOptions;
  baseUrl?: string;
  transformGetData?: (any) => string;
  httpCodeMessage?: IHttpCodeMessage;
}

interface IHttpCodeMessage {
  [key: number]: string;
}

class AcmeFetch extends Fetch {
  /**
   * 请求的实例
   */
  private instance: any = null;
  /**
   * 默认传给后端的参数（比如说token、uuid等）
   */
  private defaultParams: any = {};
  /**
   * 默认的请求配置项
   */
  private defaultOptions: IOptions = {
    method: 'GET',
    mode: 'cors',
    headers: {
      'Content-Type': 'application/json;charset=UTF-8'
    },
    credentials: 'include',
    cache: false,
    canAbort: false
  };
  /**
   * 基础请求地址
   */
  private baseUrl = 'http://localhost';
  /**
   * 缓存数据的队列
   */
  private cacheDataQueue = [];
  private cacheUrlQueue = [];
  private limit = 10;

  /**
   * 取消请求对应的Abort contronller
   */
  private controllerMap = new Map();

  /**
   * http请求默认的错误码对应的提示信息
   */
  private httpCodeMessage: IHttpCodeMessage = {
    200: '服务器成功返回请求的数据。',
    201: '新建或修改数据成功。',
    202: '一个请求已经进入后台排队（异步任务）。',
    204: '删除数据成功。',
    400: '发出的请求有错误，服务器没有进行新建或修改数据的操作。',
    401: '用户没有权限（令牌、用户名、密码错误）。',
    403: '用户得到授权，但是访问是被禁止的。',
    404: '发出的请求针对的是不存在的记录，服务器没有进行操作。',
    406: '请求的格式不可得。',
    410: '请求的资源被永久删除，且不会再得到的。',
    422: '当创建一个对象时，发生一个验证错误。',
    500: '服务器发生错误，请检查服务器。',
    502: '网关错误。',
    503: '服务不可用，服务器暂时过载或维护。',
    504: '网关超时。'
  };

  constructor({
    defalutParams,
    defaultOptions,
    baseUrl,
    transformGetData,
    httpCodeMessage
  }: IConstructor) {
    super();
    this.instance = fetch;
    this.defaultParams = defalutParams;
    this.baseUrl = baseUrl;
    this.defaultOptions = { ...this.defaultOptions, ...defaultOptions };
    this.transformGetData = transformGetData || this.transformGetData;
    this.httpCodeMessage = httpCodeMessage || this.httpCodeMessage;
  }

  /**
   * 处理为get格式的请求参数
   * @param data 需要传递给后端的数据对象
   */
  private transformGetData = (data: any) => {
    if (!data || Object.keys(data).length === 0) {
      return '';
    }
    let paramStr = '?';
    for (const key in data) {
      paramStr += `${key}=${data[key]}&`;
    }
    paramStr = paramStr.slice(0, paramStr.length - 1);
    return paramStr;
  };

  /**
   * 默认处理接口返回的json数据
   * @param data 后端返回的json数据
   */
  private defaultHandleResponse = data => {
    if (data.code === 200 || data.success) {
      if (data.result) {
        return data.result;
      }
      return null;
    }
    if (data.success === false) {
      return Promise.reject(new Error(data.message || '请求错误'));
    }
    return data;
  };

  /**
   * 对接口的返回进行状态检查
   * @param response response对象
   */
  private checkResponse = response => {
    if (response.status >= 200 && response.status < 300) {
      return response;
    } else {
      return Promise.reject(new Error(this.httpCodeMessage[response.status] || '请求异常'));
    }
  };

  /**
   * 为了达到timout超时的目的，增加一个timeout Promise
   * @param timeout 超时时间
   */
  private createTimeoutPromise = (timeout: number) => {
    return new Promise(reslove => {
      setTimeout(() => {
        reslove(
          new Response('timeout', {
            status: 504,
            statusText: 'timeout'
          })
        );
      }, timeout);
    });
  };

  /**
   * 创建一个请求，基础API
   * @param url 请求地址
   * @param data 请求body
   * @param options 请求options
   */
  public createRequest(url, data: any = {}, options: IOptions = {}) {
    let reqUrl = this.baseUrl + url;
    let controller = null;
    if (options.method === 'GET') {
      // 如果是get请求需要进行参数转换
      const paramsStr = this.transformGetData({ ...this.defaultParams, ...data });
      reqUrl = reqUrl + paramsStr;
    } else if (
      (options.dataType && options.dataType === 'json') ||
      (!options.dataType && this.defaultOptions.dataType === 'json')
    ) {
      // 如果没有传入options或者options.headers['Content-Type']没有自定义那就默认转为json格式参数
      data = JSON.stringify(data);
    }

    options.body = data;
    //GET或者HEAD请求是不能有body体的哦
    if (options.method) {
      if (options.method === 'GET' || options.method === 'HEAD') {
        delete options.body;
      }
    } else {
      if (this.defaultOptions.method === 'GET' || this.defaultOptions.method === 'HEAD') {
        delete options.body;
      }
    }
    // 设置了超时时间或者canAbort为true的话都需要去创建AbortController对象
    if (
      options.canAbort ||
      this.defaultOptions.canAbort ||
      options.timeout ||
      this.defaultOptions.timeout
    ) {
      controller = new AbortController();
      options.signal = controller.signal;
      this.controllerMap.set(url, controller);
    }
    const reqOptions = {
      ...JSON.parse(JSON.stringify(this.defaultOptions)),
      ...options
    };
    reqOptions.noContentType && delete reqOptions.headers['Content-Type'];
    // 设置了超时时间
    if (reqOptions.timeout) {
      const timeout = reqOptions.timeout;
      return Promise.race([
        this.instance(reqUrl, reqOptions),
        this.createTimeoutPromise(timeout)
      ]).then(response => {
        if (response.statusText === 'timeout' && response.status === 504) {
          this.getAortController(url).abort();
        }
        return this.checkResponse(response);
      });
    } else {
      // 正常操作返回检查过httpStatus的Resonpse
      return this.instance(reqUrl, reqOptions).then(response => {
        if (reqOptions.signal) {
          this.controllerMap.delete(url);
        }
        return this.checkResponse(response);
      });
    }
  }

  /**
   * get请求返回的是response
   * @param url 请求地址
   * @param data 传给后端的数据
   * @param options 请求方法的配置项
   */
  public get(url: string, data?: any, options?: IOptions): Promise<any> {
    return this.createRequest(url, data, { ...options, method: 'GET' });
  }

  /**
   * get请求返回的是defaultHandleResponse处理后的结果(前提是json格式的返回结果)
   * @param url 请求地址
   * @param data 传给后端的数据
   * @param options 请求方法的配置项
   */
  public getData(url: string, data?: any, options?: IOptions): Promise<any> {
    return this.get(url, data, options)
      .then(res => {
        return res.json();
      })
      .then(result => {
        const finallyRes = this.defaultHandleResponse(result);
        if (options && options.cache) {
          this.addCacheData(url, finallyRes);
        }
        return finallyRes;
      });
  }

  /**
   * post请求返回的是response
   * @param url 请求地址
   * @param data 传给后端的数据
   * @param options 请求方法的配置项
   */
  public post(url: string, data?: any, options?: IOptions): Promise<any> {
    return this.createRequest(url, data, { ...options, method: 'POST' });
  }

  /**
   * post请求回的是defaultHandleResponse处理后的结果(前提是json格式的返回结果)
   * @param url 请求地址
   * @param data 传给后端的数据
   * @param options 请求方法的配置项
   */
  public postData(url: string, data?: any, options?: IOptions): Promise<any> {
    return this.post(url, data, options)
      .then(res => {
        return res.json();
      })
      .then(result => {
        const finallyRes = this.defaultHandleResponse(result);
        if (options && options.cache) {
          this.addCacheData(url, finallyRes);
        }
        return finallyRes;
      });
  }

  /**
   * postFormData: 请求传递formdata的格式数据
   * @param url 请求地址
   * @param data 传给后端的数据
   * @param options 请求方法的配置项
   */
  public postFormData(url: string, data?: any, options?: IOptions): Promise<any> {
    if (options) {
      options.noContentType = true;
    } else {
      options = { noContentType: true };
    }
    return this.post(url, data, options)
      .then(res => {
        return res.json();
      })
      .then(result => {
        const finallyRes = this.defaultHandleResponse(result);
        if (options && options.cache) {
          this.addCacheData(url, finallyRes);
        }
        return finallyRes;
      });
  }

  /**
   * 增加cache的数据
   * @param url 请求的url也就是缓存对应的key
   * @param data 需要缓存的数据
   */
  public addCacheData(url: string, data?: any) {
    if (!this.cacheUrlQueue.includes(url)) {
      if (this.cacheDataQueue.length < this.limit) {
        this.cacheUrlQueue.push(url);
        this.cacheDataQueue.push(data);
      } else {
        this.cacheDataQueue.pop();
        this.cacheUrlQueue.pop();
        this.addCacheData(url, data);
      }
    }
  }

  /**
   * 获取cache的数据
   * @param url 请求的url也就是缓存对应的key
   */
  public getCacheData(url: string) {
    const findIndex = this.cacheUrlQueue.indexOf(url);
    if (findIndex !== -1) {
      if (findIndex !== 0) {
        const moveData = this.cacheDataQueue.splice(findIndex, 1);
        const moveUrl = this.cacheUrlQueue.splice(findIndex, 1);
        this.cacheDataQueue.unshift(moveData[0]);
        this.cacheUrlQueue.unshift(moveUrl[0]);
      }
      return this.cacheDataQueue[0];
    } else {
      console.warn('没有可使用的缓存数据');
      return false;
    }
  }

  /**
   * 获取缓存的队列信息（如果你有持久化的需求）
   */
  public getCacheQueue(): [Array<any>, Array<any>] {
    return [this.cacheDataQueue, this.cacheUrlQueue];
  }

  /**
   * 手动设置缓存队列的信息
   * @param cacheDataQueue 缓存的data队列
   * @param cacheUrlQueue 缓存的url队列
   */
  public setCacheQueue(cacheDataQueue: Array<any>, cacheUrlQueue: Array<any>) {
    if (!Array.isArray(cacheDataQueue) || !Array.isArray(cacheUrlQueue)) {
      console.error('setCacheQueue datasource type must be Array');
      return;
    }
    if (this.cacheUrlQueue.length !== this.cacheDataQueue.length) {
      console.error('setCacheQueue datasource error, check datasource');
      return;
    }
    this.cacheUrlQueue = cacheUrlQueue;
    this.cacheDataQueue = cacheDataQueue;
  }

  /**
   * 设置缓存队列的长度
   * @param {number} limit 队列的长度
   */
  public setLimit(limit: number) {
    if (typeof limit === 'number') {
      this.limit = limit;
    } else {
      console.error('limit 非法');
    }
  }

  /**
   * 获取请求的AortController可以进行请求abort操作
   * @param key 请求的url
   */
  public getAortController(key: string) {
    if (this.controllerMap.has(key)) {
      return this.controllerMap.get(key);
    } else {
      // 兼容当fetch进入了reslove状态后清除了AbortController或者根本没有此对象，默认给个空方法不会报错；
      return {
        abort: () => {}
      };
    }
  }

  /**
   * 清除我们的缓存队列
   */
  public clearCache() {
    this.cacheDataQueue = [];
    this.cacheUrlQueue = [];
  }

  /**
   * 清除我们的缓存队列以及缓存的abortContronller
   */
  public clearAll() {
    this.cacheDataQueue = [];
    this.cacheUrlQueue = [];
    this.controllerMap = new Map();
  }
}

/**
 * cover 住我们的await prmise场景
 */
const useAwaitPromise = promise => {
  if (!promise || !Promise.prototype.isPrototypeOf(promise)) {
    return new Promise((resolve, reject) => {
      reject(new Error('useAwaitPromise方法参数需要是一个promise对象'));
    }).catch(err => {
      return [err, null];
    });
  }
  return promise
    .then(function (data) {
      return [null, data];
    })
    .catch(err => {
      return [err, null];
    });
};

export { useAwaitPromise };
export default AcmeFetch;
